/*************************************************************************/ /*!
@File			innodpu_vga.c
@Title
@Copyright		Copyright (c) Innosilicon Technology Ltd. All Rights Reserved
@Description
@License		Dual MIT/GPLv2

The contents of this file are subject to the MIT license as set out below.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

Alternatively, the contents of this file may be used under the terms of
the GNU General Public License Version 2 ("GPL") in which case the provisions
of GPL are applicable instead of those above.

If you wish to allow use of your version of this file only under the terms of
GPL, and not to allow others to use your version of this file under the terms
of the MIT license, indicate your decision by deleting the provisions above
and replace them with the notice and other provisions required by GPL as set
out in the file called "GPL-COPYING" included in this distribution. If you do
not delete the provisions above, a recipient may use your version of this file
under the terms of either the MIT license or GPL.

This License is also included in this distribution in the file called
"MIT-COPYING".

EXCEPT AS OTHERWISE STATED IN A NEGOTIATED AGREEMENT: (A) THE SOFTWARE IS
PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE AND NONINFRINGEMENT; AND (B) IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/ /**************************************************************************/

#include <linux/platform_device.h>
#include "innodpu_vga.h"
#include "innodpu_connector.h"
#include "innodpu_vga_debugfs.h"
#include "innopmbus_drv.h"

static int vga_ddc_transfer_size __read_mostly = I2C_SMBUS_BLOCK_MAX;
module_param_unsafe(vga_ddc_transfer_size, int, 0644);
MODULE_PARM_DESC(vga_ddc_transfer_size,
		"Number of bytes to transfer in a single I2C over VGA DDCCI message, (default max 32)");
static bool s_vga_ddcci = true;
module_param(s_vga_ddcci, bool, 0600);
MODULE_PARM_DESC(s_vga_ddcci, "vga ddcci-pmbus (default: false)");

static struct innodpu_mode_bypass_list s_vga_mode_bypass_list[] =
{
	// {"*", "JTY", CHIP_G0M_SOC, 1280, 800, 60},
};

static const struct drm_display_mode s_vga_noedid_modes[] = {
	{DRM_MODE("1920x1080", DRM_MODE_TYPE_DRIVER, 148500, 1920, 2008,
			  2052, 2200, 0, 1080, 1084, 1089, 1125, 0,
			  DRM_MODE_FLAG_PHSYNC | DRM_MODE_FLAG_PVSYNC),
	 .picture_aspect_ratio = HDMI_PICTURE_ASPECT_16_9,},
};

static RAW_NOTIFIER_HEAD(vga_notifier);
void vga_notifier_register(struct notifier_block *nb)
{
	raw_notifier_chain_register(&vga_notifier, nb);
}
void vga_notifier_unregister(struct notifier_block *nb)
{
	raw_notifier_chain_unregister(&vga_notifier, nb);
}

static int inno_vga_without_edid(struct drm_connector *connector)
{
	int index = 0;
	int num_modes = 0;
	const struct drm_display_mode *ptr = NULL;
	struct drm_display_mode *mode = NULL;
	struct drm_device *drm_dev = connector->dev;

	for (index = 0, num_modes = 0; index < ARRAY_SIZE(s_vga_noedid_modes); ++index) {
		ptr = &s_vga_noedid_modes[index];
		if ((ptr->hdisplay > drm_dev->mode_config.max_width) ||
				(ptr->vdisplay > drm_dev->mode_config.max_height)) {
			continue;
		}

		if (drm_mode_vrefresh(ptr) > (INNOVGA_DEFAULT_FRAME_RATE + 1)) {
			continue;
		}

		mode = drm_mode_duplicate(connector->dev, ptr);
		if (mode) {
			drm_mode_probed_add(connector, mode);
			++num_modes;
		}
	}
	return num_modes;
}

static enum drm_mode_status inno_vga_encoder_mode_valid(
		struct drm_encoder *encoder, const struct drm_display_mode *mode)
{
	struct vga_device_t *inno_vga = to_inno_vga(encoder);

	innodpu_info(inno_vga->dev, DPU_UT_VGA,
		"%s encoder mode valid resolution: %dx%d@%d",
		inno_vga->name, mode->hdisplay, mode->vdisplay, drm_mode_vrefresh(mode));

	if (inno_vga->chip.encoder_mode_valid)
		return inno_vga->chip.encoder_mode_valid(&inno_vga->chip, mode);

	return MODE_OK;
}

bool inno_vga_mode_is_push(struct vga_device_t *inno_vga, struct drm_display_mode *adjusted_mode)
{
	// TBD: RK??? not support
	if ( (adjusted_mode->htotal == inno_vga->native_mode.htotal) &&
		(adjusted_mode->vtotal == inno_vga->native_mode.vtotal) &&
		(adjusted_mode->clock == inno_vga->native_mode.clock) &&
		((adjusted_mode->hdisplay != inno_vga->native_mode.hdisplay)||
		(adjusted_mode->vdisplay != inno_vga->native_mode.vdisplay)) ) {
		return true;
	}

	return false;
}

bool inno_vga_native_mode_valid(struct vga_device_t *inno_vga)
{
	const struct drm_display_mode *mode = &inno_vga->native_mode;
	if ((mode->clock > 0) &&  ((mode->hdisplay > 0) && (mode->hdisplay < 8192)))
		return true;

	return false;
}

static bool inno_vga_encoder_mode_fixup(struct drm_encoder *encoder,
		const struct drm_display_mode *mode, struct drm_display_mode *adjusted_mode)
{
	struct vga_device_t *inno_vga
		= container_of(encoder, struct vga_device_t, encoder);

	if (inno_vga_native_mode_valid(inno_vga) && inno_vga_mode_is_push(inno_vga, adjusted_mode)) {
		drm_mode_copy(adjusted_mode, &inno_vga->native_mode);
	}
	return true;
}

static void inno_vga_auto_setup_mode_enable(struct vga_device_t *inno_vga,
		struct drm_display_mode *adjusted_mode)
{
	struct drm_display_mode *mode = NULL;
	struct drm_display_mode *autoset_mode = NULL;

	if (!inno_vga || !adjusted_mode)
		return;

	inno_mutex_lock(inno_vga->auto_setup_mutex);

	inno_vga->auto_setup_enable = true;

	if (!s_vga_adpt_every_mode) {
		list_for_each_entry(mode, &inno_vga->auto_setup_list, head) {
			if (mode && drm_mode_equal(mode, adjusted_mode)) {
				inno_vga->auto_setup_enable = false;
				break;
			} else {
				inno_vga->auto_setup_enable = true;
			}
		}
	}

	innodpu_info(inno_vga->dev, DPU_UT_VGA,
			"AutoSetupEnable:%d", inno_vga->auto_setup_enable);

	if (inno_vga->auto_setup_enable) {
		autoset_mode = drm_mode_duplicate(inno_vga->connector.dev, adjusted_mode);
		if (autoset_mode)
			list_add_tail(&autoset_mode->head, &inno_vga->auto_setup_list);
	}

	inno_mutex_unlock(inno_vga->auto_setup_mutex);
}

static void inno_vga_auto_setup_mode_clear(struct vga_device_t *inno_vga)
{
	struct drm_display_mode *mode = NULL;
	struct drm_display_mode *mode1 = NULL;

	if (!inno_vga)
		return;

	innodpu_info(inno_vga->dev, DPU_UT_VGA, "AutoSetupModeClear");

	inno_mutex_lock(inno_vga->auto_setup_mutex);

	list_for_each_entry_safe(mode, mode1, &inno_vga->auto_setup_list, head) {
		list_del(&mode->head);
		drm_mode_destroy(inno_vga->connector.dev, mode);
	}

	inno_vga->auto_setup_enable = false;

	inno_mutex_unlock(inno_vga->auto_setup_mutex);
}

static int inno_vga_set_vcp_feature(struct vga_device_t *vga, int vcp_code, int value)
{
	struct i2c_msg msg;
	char buf[7] = {0x51, 0x84, 0x03,};
	int i = 0;

	innodpu_info(vga->dev, DPU_UT_VGA,
			"set vcp-%d value-%d\n", vcp_code, value);

	buf[3] = vcp_code;
	buf[4] = (value >> 8) & 0xff;
	buf[5] = value & 0xff;
	buf[6] = DDC_CI_ADDR * 2;

	for (i = 0; i < (ARRAY_SIZE(buf) - 1); i++)
		buf[6] ^= buf[i];

	msg.addr = DDC_CI_ADDR;
	msg.flags = 0;
	msg.len = ARRAY_SIZE(buf);
	msg.buf = buf;
	if (i2c_transfer(&vga->ddc, &msg, 1) != 1)
		return -EFAULT;

	return 0;
}

int inno_vga_auto_setup_ctrl(struct vga_device_t *vga, bool enable)
{
	innodpu_info(vga->dev, DPU_UT_VGA,
			"auto_setup_ctrl:%s\n", enable ? "on" : "off");

	return inno_vga_set_vcp_feature(vga, 0xa2, enable ? 0x2 : 0x1);
}

int inno_vga_auto_setup(struct vga_device_t *vga, bool active)
{
	innodpu_info(vga->dev, DPU_UT_VGA,
			"auto_setup:%s\n", active ? "active" : "not active");

	return inno_vga_set_vcp_feature(vga, 0x1e, active ? 0x1 : 0x0);
}

static void inno_vga_auto_setup_execute(struct work_struct *work)
{
	struct delayed_work *auto_setup_work = to_delayed_work(work);
	struct vga_device_t *inno_vga = to_inno_vga(auto_setup_work);

	if (!inno_vga)
		return;

	if (inno_vga->auto_setup_enable) {
		inno_mutex_lock(inno_vga->adapt_lock);
		atomic_set(&inno_vga->adapt_cnt, 1);

		/* show white edge here */
		raw_notifier_call_chain(&vga_notifier, true, &inno_vga->current_mode);
		inno_vga_auto_setup_ctrl(inno_vga, true);
		inno_usleep_range(50000, 60000);
		inno_vga_auto_setup(inno_vga, true);

		/* delay 5 seconds to wait vga auto adapt finished */
		inno_wait_event_interruptible_timeout(inno_vga->adapt_wait,
			atomic_read(&inno_vga->adapt_cnt) == 0, msecs_to_jiffies(5000));

		/* close white edge here */
		raw_notifier_call_chain(&vga_notifier, false, &inno_vga->current_mode);
		inno_mutex_unlock(inno_vga->adapt_lock);
	}
}

static void inno_vga_encoder_mode_set(struct drm_encoder *encoder,
		struct drm_display_mode *mode, struct drm_display_mode *adjusted_mode)
{
	struct vga_device_t *inno_vga = to_inno_vga(encoder);
	int dpu_id = 0;

	if (!encoder || !adjusted_mode) {
		return;
	}

	dpu_id = innodpu_get_dpuid_bycrtc(encoder->crtc);

	if (inno_vga->chip.encoder_modeset) {
		innodpu_info(inno_vga->dev, DPU_UT_VGA,"%s modeset: "DRM_MODE_FMT "\n",
			inno_vga->name, DRM_MODE_ARG(adjusted_mode));
		inno_vga->chip.encoder_modeset(&inno_vga->chip, dpu_id, inno_vga->chip.test_mode, adjusted_mode);
		drm_mode_copy(&inno_vga->current_mode,  adjusted_mode);
	}

	inno_vga_auto_setup_mode_enable(inno_vga, adjusted_mode);
}

static void inno_vga_encoder_mode_disable(struct drm_encoder *encoder)
{
	struct vga_device_t *inno_vga = to_inno_vga(encoder);

	if (NULL == encoder) {
		return;
	}

	if (encoder->crtc) {
		innodpu_info(inno_vga->dev, DPU_UT_VGA,
			"inno %s encoder mode disable\n",inno_vga->name );
	} else {
		innodpu_info(inno_vga->dev, DPU_UT_VGA,
			"inno %s encoder mode disable, crtc is null\n",inno_vga->name);
	}

	if (inno_vga->chip.encoder_disable)
		inno_vga->chip.encoder_disable(&inno_vga->chip);

	atomic_set(&inno_vga->adapt_cnt, 0);
	inno_wake_up_interruptible(inno_vga->adapt_wait);
	cancel_delayed_work_sync(&inno_vga->auto_setup_work);
}

static void inno_vga_encoder_mode_enable(struct drm_encoder *encoder)
{
	struct vga_device_t *inno_vga = to_inno_vga(encoder);

	if (NULL == encoder) {
		return;
	}

	if (encoder->crtc) {
		innodpu_info(inno_vga->dev, DPU_UT_VGA,
			"inno %s encoder mode enable\n",inno_vga->name );
	} else {
		innodpu_info(inno_vga->dev, DPU_UT_VGA,
			"inno %s encoder mode enable, crtc is null\n",inno_vga->name);
	}

	if (inno_vga->chip.encoder_enable)
		inno_vga->chip.encoder_enable(&inno_vga->chip, encoder->crtc);

	if (s_vga_auto_adapt)
		queue_delayed_work(inno_vga->auto_setup_wq, &inno_vga->auto_setup_work, msecs_to_jiffies(3000));
}

static int inno_vga_encoder_destroy(struct drm_encoder *encoder)
{
	BUG_ON(encoder);

	/*
	 * Called  drm_encoder_cleanup
	 */
	if (encoder->funcs && encoder->funcs->destroy)
		encoder->funcs->destroy(encoder);

	return 0;
}

static const struct drm_encoder_funcs s_inno_vga_encoder_funcs = {
	.destroy = drm_encoder_cleanup,
};

static const struct drm_encoder_helper_funcs s_inno_vga_encoder_helper_funcs = {
	.mode_fixup = inno_vga_encoder_mode_fixup,
	.mode_valid = inno_vga_encoder_mode_valid,
	.mode_set = inno_vga_encoder_mode_set,
	.disable = inno_vga_encoder_mode_disable,
	.enable = inno_vga_encoder_mode_enable,
};

static int inno_vga_encoder_create(struct drm_device *drm_dev,
		struct drm_encoder *encoder, unsigned int possible_crtc)
{
	int retcode = 0;

	drm_encoder_helper_add(encoder, &s_inno_vga_encoder_helper_funcs);
	retcode = drm_encoder_init(drm_dev, encoder,
		&s_inno_vga_encoder_funcs, DRM_MODE_ENCODER_DAC, NULL);

	encoder->possible_crtcs = possible_crtc;

	return retcode;
}

static void inno_vga_connector_destroy_func(struct drm_connector *connector)
{
	if (!connector) {
		return;
	}

	drm_connector_unregister(connector);
	drm_connector_cleanup(connector);
}

static s32 inno_vga_get_edid_block(void *data, u8 * buf, u32 block, size_t len)
{
	struct vga_chip_t *chip = (struct vga_chip_t*)data;

	if (len > INNOVGA_EDID_BUF_LEN / 2) {
		return -EINVAL;
	}

	if (block % 2 == 0) {
		inno_memcpy(buf, chip->edid_buf, len);
	} else {
		inno_memcpy(buf, chip->edid_buf + INNOVGA_EDID_BUF_LEN / 2, len);
	}

	return 0;
}

static int inno_vga_get_modes_strpush(struct drm_connector *connector)
{
	int ret = 0;
	struct vga_device_t *vga_dev = to_inno_vga(connector);
	struct vga_chip_t *chip = &vga_dev->chip;

	ret = hal_vga_edid_data(inno_dev_get_parent(chip->dev),
			vga_dev->vga_id, chip->edid_buf);
	if (ret) {
		innodpu_err(chip->dev, "strpush edid error, ret:%d\n", ret);
		return -EFAULT;
	}

	chip->modes = innodpu_str_push_edid(chip->edid_buf, connector);
	innodpu_info(vga_dev->dev, DPU_UT_VGA, "get edid from strpush ok\n");
	return ret;
}

static int inno_vga_get_edid_user(struct drm_connector *connector)
{
	int ret = 0;
	int i = 0;
	struct vga_device_t *vga_dev = to_inno_vga(connector);
	struct vga_chip_t *chip = &vga_dev->chip;

	ret = hal_vga_edid_data(inno_dev_get_parent(chip->dev),
			vga_dev->vga_id, chip->edid_buf);
	if (ret) {
		innodpu_err(chip->dev, "get edid from user error, ret:%d\n", ret);
		return -EFAULT;
	}

	for (i = 0; i <= chip->edid_buf[0x7e] && i <= 1; i++) {
		if (!drm_edid_block_valid(chip->edid_buf + i * INNOVGA_EDID_BUF_LEN / 2, i, false, NULL)) {
			ret = -EFAULT;
			break;
		}
	}
	innodpu_info(vga_dev->dev, DPU_UT_VGA, "get edid from user ok\n");
	return ret;
}

static int inno_vga_read_edid(struct vga_device_t *vga)
{
	struct i2c_msg msg;
	int ret = 0;

	msg.addr = DDC_EDID_ADDR;
	msg.flags = I2C_M_RD;
	msg.len = INNOVGA_EDID_BUF_LEN / 2;
	msg.buf = vga->chip.edid_buf;
	inno_memset(vga->chip.edid_buf, 0, sizeof(vga->chip.edid_buf));
	innodpu_info(vga->dev, DPU_UT_VGA, "s_vga_ddcci =%d \n", s_vga_ddcci);
	if (s_vga_ddcci) {
		if (i2c_transfer(&vga->ddc, &msg, 1) != 1)
			return -EFAULT;
	} else if (vga->chip.connector_get_edid) {
		if (vga->chip.connector_get_edid(&vga->chip))
			return -EFAULT;
	} else {
		innodpu_err(vga->dev, "Not support %s chip detect handle !!!\n", vga->name);
		return -EFAULT;
	}

	if (drm_edid_header_is_valid(vga->chip.edid_buf) != 8)
		return -EFAULT;
	innodpu_info(vga->dev, DPU_UT_VGA, "get edid from monitor ok\n");
	return ret;
}


static int inno_vga_get_edid(struct drm_connector *connector)
{
	int ret = 0;
	struct vga_device_t *vga_dev =  to_inno_vga(connector);
	struct vga_chip_t *chip = &vga_dev->chip;

	inno_memset(chip->edid_buf, 0, sizeof(chip->edid_buf));

	switch (chip->hal_edid_mode) {
	case EDID_AUTO_READ:
		ret = inno_vga_read_edid(vga_dev);
	break;
	case EDID_STR_PUSH:
		ret = inno_vga_get_modes_strpush(connector);
	break;
	case EDID_USER_DEFINE:
		ret = inno_vga_get_edid_user(connector);
	break;
	default:
		innodpu_err(vga_dev->dev, "get_edid failed, invalid hal_edid_mode,"
				"read edid from monitor\n");
		ret = -EFAULT;
	}
	innodpu_info(vga_dev->dev, DPU_UT_VGA,  "hal_edid_mode:%d ret = %d\n", chip->hal_edid_mode, ret);

	if (ret == 0 && chip->hal_edid_mode != EDID_STR_PUSH) {
		vga_dev->chip.vga_edid = drm_do_get_edid(connector, inno_vga_get_edid_block, &vga_dev->chip);
		chip->max_clk = innodpu_conn_get_monitor_max_clk((u8 *)chip->vga_edid);
		if (chip->max_clk <= 0) {
			chip->max_clk = 175000;
			innodpu_info(vga_dev->dev, DPU_UT_VGA, "get vga max clock err, set it to 175MHz\n");
		}
	}
	if (ret || (chip->hal_edid_mode != EDID_STR_PUSH && vga_dev->chip.vga_edid == NULL)) {
		if (vga_dev->chip.vga_edid) {
			inno_kfree(vga_dev->chip.vga_edid);
			vga_dev->chip.vga_edid = NULL;
		}
		innodpu_err(vga_dev->dev, "[%s] Edid Invalid, may be poor contact, please re-plug\n", connector->name);
		drm_connector_update_edid_property(connector, NULL);
		vga_dev->chip.modes = inno_vga_without_edid(connector);

	} else if (chip->hal_edid_mode != EDID_STR_PUSH) {
		drm_connector_update_edid_property(connector, vga_dev->chip.vga_edid);
		vga_dev->chip.modes = drm_add_edid_modes(connector, vga_dev->chip.vga_edid);
	}

	return ret;
}

static bool inno_vga_native_mode_filter(struct drm_connector *connector,
	struct drm_display_mode *mode, bool priv_filter)
{
	struct vga_device_t *inno_vga
		= container_of(connector, struct vga_device_t, connector);
	struct drm_encoder *encoder = &inno_vga->encoder;
	int ret = 0;

	// TBD: current not support 4k scaler
	if (mode && priv_filter && ((mode->hdisplay > INNODPU_COMBINE_WIDTH) ||
		(mode->vdisplay > INNODPU_COMBINE_HEIGHT)) )
		return true;

	if (mode && mode->clock > inno_vga->chip.max_clk)
		return true;

	if (connector->helper_private && connector->helper_private->mode_valid) {
		ret = connector->helper_private->mode_valid(&inno_vga->connector, mode);
		if (ret != MODE_OK) {
			return true;
		}
	}

	if (encoder->helper_private && encoder->helper_private->mode_valid) {
		ret = encoder->helper_private->mode_valid(&inno_vga->encoder, mode);
		if (ret != MODE_OK) {
			return true;
		}
	}

	return false;
}

/**
 * inno_vga_connector_helper_get_modes - Get vga supports mode;
 *
 * Returns:
 * Number of supported modes
 */
static int inno_vga_connector_helper_get_modes(
		struct drm_connector *connector)
{
	struct drm_display_mode *mode = NULL;
	struct drm_display_mode *mode1 = NULL;
	struct vga_device_t *inno_vga = to_inno_vga(connector);
	struct edid *edid = inno_vga->chip.vga_edid;
	struct drm_display_mode *native_mode;
	int hdisplay, vdisplay;
	char mode_name[DRM_DISPLAY_MODE_LEN];

	inno_memset(&inno_vga->native_mode, 0, sizeof(inno_vga->native_mode));

	if (drm_edid_is_valid(edid)) {
		drm_connector_update_edid_property(connector, edid);
		inno_vga->chip.modes = drm_add_edid_modes(connector, edid);
	} else {
		inno_vga_get_edid(connector);
	}

	// get native mode
	drm_mode_sort(&connector->probed_modes);
	native_mode = innodpu_get_native_mode(connector, inno_vga_native_mode_filter);

	if (native_mode) {
		list_for_each_entry_safe(mode, mode1, &connector->probed_modes, head) {
			if (!mode) {
				continue;
			}
			if (mode->hdisplay <= 1600 &&
					mode->hdisplay <= native_mode->hdisplay &&
						mode->vdisplay <= native_mode->vdisplay) {
				hdisplay = mode->hdisplay;
				vdisplay = mode->vdisplay;
				drm_mode_copy(mode, native_mode);
				snprintf(mode_name, sizeof(mode_name), "%dx%d", hdisplay, vdisplay);
				strncpy(mode->name, mode_name, DRM_DISPLAY_MODE_LEN);
				mode->hdisplay = hdisplay;
				mode->vdisplay = vdisplay;
				mode->type &= ~DRM_MODE_TYPE_PREFERRED;
			}
		}
	}

	// add common mode
	if (native_mode) {
		innodpu_info(inno_vga->dev, DPU_UT_VGA, "%s native mode: "DRM_MODE_FMT ", status:%d\n",
			inno_vga->name, DRM_MODE_ARG(native_mode), native_mode->status);
		inno_memcpy(&inno_vga->native_mode, native_mode, sizeof(inno_vga->native_mode));
		drm_mode_set_crtcinfo(&inno_vga->native_mode, CRTC_INTERLACE_HALVE_V);
		inno_vga->native_mode.status = MODE_OK;
		inno_vga->chip.modes += innodpu_connector_add_common_modes(connector, &inno_vga->native_mode, true);
	}

	return inno_vga->chip.modes;
}


/**
 * inno_vga_connector_detect_ctx - check vga hpd status
 *
 * Returns:
 * drm_connector_status. (connector_status_connected or connector_status_disconnected)
 */
static int inno_vga_connector_detect_ctx(struct drm_connector *connector,
		struct drm_modeset_acquire_ctx *ctx, bool force)
{
	int ret = connector_status_disconnected;
	struct vga_device_t *inno_vga = to_inno_vga(connector);

	atomic64_set(&inno_vga->chip.hpg_detect, 1);

	/*Use pmbus to simulate i2c reading edid to perform
	 * hotplug detection*/
	if (s_vga_ddcci) {
		if (inno_vga_read_edid(inno_vga) == 0) {
			ret = connector_status_connected;
		}
	} else if (inno_vga->chip.connector_detect) {
		ret = inno_vga->chip.connector_detect(&inno_vga->chip);
	} else {
		innodpu_warn(inno_vga->dev, "Not support %s chip detect handle !!!\n",
				inno_vga->name);
	}

	if (ret != connector_status_connected) {
		if (inno_vga->chip.vga_edid) {
			kfree(inno_vga->chip.vga_edid);
			inno_vga->chip.vga_edid = NULL;
		}
		inno_memset(&inno_vga->native_mode, 0, sizeof(inno_vga->native_mode));

		if (!list_empty(&inno_vga->auto_setup_list))
			inno_vga_auto_setup_mode_clear(inno_vga);
	}

	atomic64_set(&inno_vga->chip.hpg_detect, 0);
	atomic64_set(&inno_vga->chip.connect_status, ret);

	innodpu_info(inno_vga->dev, DPU_UT_VGA, "connector_status: %s\n",
			(ret == connector_status_connected) ? "connect" : "disconnect");

	return ret;
}

static enum drm_mode_status inno_vga_connector_helper_mode_valid(
		struct drm_connector *connector, struct drm_display_mode *mode)
{
	int status = MODE_OK;
	struct vga_device_t *inno_vga = to_inno_vga(connector);

	if (mode->flags & DRM_MODE_FLAG_DBLSCAN) {
		innodpu_info(inno_vga->dev, DPU_UT_VGA, "[Connector:%d] Doublescan mode rejected.\n", connector->base.id);
		return MODE_NO_DBLESCAN;
	}

	if (mode->flags & DRM_MODE_FLAG_INTERLACE) {
		innodpu_info(inno_vga->dev, DPU_UT_VGA, "[Connector:%d] Interlace mode rejected.\n", connector->base.id);
		return MODE_NO_INTERLACE;
	}

	if (drm_mode_vrefresh(mode) > INNOVGA_DEFAULT_FRAME_RATE) {
		innodpu_info(inno_vga->dev, DPU_UT_VGA,
			"[Connector:%d] %s:vfresh must be smaller than 60 !\n",
			connector->base.id, mode->name);
		return MODE_BAD_HVALUE;
	}

	if (mode->hdisplay % 16) {
		return MODE_BAD_HVALUE;
	}

	if (inno_vga->chip.connector_mode_valid) {
		status = inno_vga->chip.connector_mode_valid(&inno_vga->chip, mode);
		if (status != MODE_OK)
			return status;
	}

	/* check  */
	{
		int ret = 0;
		char monitor_name[4] = {0};
		char odm_info[ODM_INFO_LEN] = {0};
		chip_type_e plat = 0;

		plat = hal_get_chiptype(inno_vga->drm_dev->dev);
		ret += innodpu_get_odm_info(inno_vga->dev, odm_info, inno_vga->drm_dev->dev);
		ret += innodpu_get_mfc_name(inno_vga->dev, monitor_name, inno_vga->chip.edid_buf);
		if (!ret) {
			if(innodpu_mode_need_bypass(monitor_name, odm_info, plat,
				s_vga_mode_bypass_list, ARRAY_SIZE(s_vga_mode_bypass_list),
				mode->hdisplay, mode->vdisplay, drm_mode_vrefresh(mode))) {
				innodpu_info(inno_vga->dev, DPU_UT_VGA, " mfc[%s] odm[%s] %dx%d@%d was filtered\n",
					monitor_name, odm_info,
					mode->hdisplay, mode->vdisplay, drm_mode_vrefresh(mode));
				return MODE_NOMODE;
			}
		}
	}

	return MODE_OK;
}

static void inno_vga_connector_print_state(struct drm_printer *p,
	const struct drm_connector_state *state)
{

}

s32 _i2c_smbus_xfer(struct i2c_adapter *adapter, u16 addr, unsigned short flags,
char read_write, u8 command, int protocol, union i2c_smbus_data *data)
{
	struct inno_pmbus_dev *pmbus = i2c_get_adapdata(adapter);
	s32 ret = 0;

	if (!pmbus) {
		inno_error("%s %d pmbus get failed\n", __func__, __LINE__);
		return -1;
	}

	mutex_lock(&pmbus->chip_rs_lock);
	pmbus->chip_need_rs = 1;
	ret = i2c_smbus_xfer(adapter, addr, flags, read_write, command, protocol, data);
	mutex_unlock(&pmbus->chip_rs_lock);
	return ret;
}

static int innovga_i2c_xfer(struct i2c_adapter *ddc, struct i2c_msg *msgs,
			int num)
{
	struct vga_device_t *inno_vga = ddc->algo_data;
	u8 request = 0;
	u8 command = 0;
	unsigned int i, j;
	unsigned transfer_size;
	unsigned int slave_addr = 0;
	union i2c_smbus_data data;

	if ((inno_vga->pmbus_adapter == NULL) ||
		s_vga_ddcci == false ||
		( atomic64_read(&inno_vga->chip.connect_status) != connector_status_connected &&
		(!atomic64_read(&inno_vga->chip.hpg_detect)))) {
		return -EFAULT;
	}

	transfer_size = vga_ddc_transfer_size;
	for (i = 0; i < num; i++) {
		request = (msgs[i].flags & I2C_M_RD) ? I2C_SMBUS_READ : I2C_SMBUS_WRITE;
		slave_addr = msgs[i].addr;
		for (j = 0; j < msgs[i].len; j += request ? data.block[0] : (data.block[0] + 1)) {
			memset(data.block, 0, sizeof(data.block));
			if (request == I2C_SMBUS_WRITE) {
				data.block[0] = min(transfer_size, msgs[i].len - j - 1);
				command = msgs[i].buf[j];
				if (data.block[0] >= 1)
					memcpy(&data.block[1], msgs[i].buf + j + 1, data.block[0]);
			} else {
				data.block[0] = min(transfer_size, msgs[i].len - j);
				command = j;
			}

			/*The delayed operation during DDC/CI communication is handled by the application layer,
			 * and the driver does not need to delay*/
			/*if (slave_addr == DDCCI_ADDR && request == I2C_SMBUS_READ)
				inno_usleep_range(50000, 60000);*/

			innodpu_info(inno_vga->dev, DPU_UT_VGA,
					"%s slave_addr:%#x command:%#.2x data.block[0]:%d msgs.len:%d\n",
					request ? "read" : "write", msgs[i].addr,
					command, data.block[0], msgs[i].len);
			_i2c_smbus_xfer(inno_vga->pmbus_adapter, slave_addr,
						0x0, request, command,
						I2C_SMBUS_I2C_BLOCK_DATA, &data);
			if (request == I2C_SMBUS_READ)
				memcpy(msgs[i].buf + j, &data.block[1], data.block[0]);

			if (data.block[0] && (s_dpu_debug & DPU_UT_VGA)) {
				print_hex_dump(KERN_NOTICE, "\t", 0, 16, 1,
					(request == I2C_SMBUS_READ) ? &msgs[i].buf[j] : &data.block[1], data.block[0], false);
			}
		}
	}

	return num;
}

static u32 innovga_i2c_functionality(struct i2c_adapter *ddc)
{
	return I2C_FUNC_I2C | I2C_FUNC_SMBUS_EMUL |
	       I2C_FUNC_SMBUS_READ_BLOCK_DATA |
	       I2C_FUNC_SMBUS_BLOCK_PROC_CALL |
	       I2C_FUNC_10BIT_ADDR;
}

static const struct i2c_algorithm innovga_i2c_algo = {
	.functionality = innovga_i2c_functionality,
	.master_xfer = innovga_i2c_xfer,
};

static int inno_vga_connector_late_register(struct drm_connector *connector)
{
	int ret = 0;
	struct device *adap_dev = NULL;
	struct vga_device_t *inno_vga = to_inno_vga(connector);

	inno_vga->ddc.class = I2C_CLASS_DDC;
	inno_vga->ddc.owner = THIS_MODULE;
	inno_vga->ddc.dev.parent = inno_vga->dev;
	inno_vga->ddc.algo_data = inno_vga;
	inno_vga->ddc.algo = &innovga_i2c_algo;
	strlcpy(inno_vga->ddc.name, dev_name(inno_vga->dev), sizeof(inno_vga->ddc.name));

	ret = i2c_add_adapter(&inno_vga->ddc);
	if (ret) {
		innodpu_err(inno_vga->dev, "vga-ddc register failed\n");
		goto end;
	}

	adap_dev = &inno_vga->ddc.dev;
	ret = sysfs_create_link(&connector->kdev->kobj,
				&adap_dev->kobj,
				adap_dev->kobj.name);
	if (ret) {
		innodpu_err(inno_vga->dev, "%s sysfs link create failed\n", inno_vga->ddc.name);
	}

	ret = inno_vga_custom_debugfs_create(connector);
	if (ret) {
		innodpu_err(inno_vga->dev, "%s custom debugfs create failed-%d\n",
			inno_vga->name, ret);
	}

end:
	return ret;
}

static void inno_vga_connector_early_unregister(struct drm_connector *connector)
{
	struct vga_device_t *inno_vga = to_inno_vga(connector);

	sysfs_remove_link(&connector->kdev->kobj,
			inno_vga->ddc.dev.kobj.name);

	i2c_del_adapter(&inno_vga->ddc);
}

static struct drm_connector_helper_funcs s_inno_vga_connector_helper_funcs = {
	.get_modes = inno_vga_connector_helper_get_modes,
	.mode_valid = inno_vga_connector_helper_mode_valid,
	.detect_ctx = inno_vga_connector_detect_ctx,
};

static const struct drm_connector_funcs s_inno_vga_connector_funcs = {
	.fill_modes = drm_helper_probe_single_connector_modes,
	.destroy = inno_vga_connector_destroy_func,
	.atomic_print_state = inno_vga_connector_print_state,
	.reset = drm_atomic_helper_connector_reset,
	.atomic_duplicate_state = drm_atomic_helper_connector_duplicate_state,
	.atomic_destroy_state = drm_atomic_helper_connector_destroy_state,
	.late_register = inno_vga_connector_late_register,
	.early_unregister = inno_vga_connector_early_unregister,
};

static int inno_vga_connector_create(struct drm_device *drm_dev,
				struct drm_connector *connector)
{
	int retcode = 0;

	BUG_ON(!drm_dev);
	BUG_ON(!connector);

	drm_connector_helper_add(connector, &s_inno_vga_connector_helper_funcs);

	retcode = drm_connector_init(drm_dev, connector, &s_inno_vga_connector_funcs,
		DRM_MODE_CONNECTOR_VGA);
	connector->polled = DRM_CONNECTOR_POLL_CONNECT | DRM_CONNECTOR_POLL_DISCONNECT;  // TBD : DRM_CONNECTOR_POLL_CONNECT ?
	connector->dpms = DRM_MODE_DPMS_OFF;
	connector->interlace_allowed = false;
	connector->doublescan_allowed = false;
	connector->display_info.subpixel_order = SubPixelHorizontalRGB;

	return retcode;
}

static int inno_vga_connector_destroy(struct drm_connector *connector)
{
	struct drm_mode_object *conn_obj = &connector->base;

	while (kref_read(&conn_obj->refcount) > 0)
		drm_connector_put(connector); // will be called drm_connector_free
	return 0;
}


static int inno_vga_connector_attach_encoder(
			struct vga_device_t *inno_vga)
{
	int retcode = 0;

	retcode = inno_vga_encoder_create(inno_vga->drm_dev, &inno_vga->encoder,
		inno_vga->chip.possible_crtc);
	if (retcode) {
		innodpu_err(inno_vga->dev, "%s encoder create failed-%d\n",
				  inno_vga->name, retcode);
		return retcode;
	}

	retcode = inno_vga_connector_create(inno_vga->drm_dev, &inno_vga->connector);
	if (retcode) {
		innodpu_err(inno_vga->dev, "%s connector create failed-%d\n",
			inno_vga->name, retcode);
		goto err_connector_create;
	}

	// only set connector->possible_encoder,
	retcode = drm_connector_attach_encoder(&inno_vga->connector, &inno_vga->encoder);
	if (retcode) {
		innodpu_err(inno_vga->dev, "%s connector attach encoder failed-%d\n",
			inno_vga->name, retcode);
		goto error_attach_failed;
	}

	innodpu_info(inno_vga->dev, DPU_UT_VGA, "%s attach encoder-%d connector-%d\n",
		inno_vga->name, inno_vga->encoder.base.id, inno_vga->connector.base.id);

	return retcode;

error_attach_failed:
	inno_vga_connector_destroy(&inno_vga->connector);
err_connector_create:
	inno_vga_encoder_destroy(&inno_vga->encoder);

	return retcode;
}

void inno_vga_hw_irq_enable(struct vga_device_t *inno_vga, unsigned int flags)
{
	BUG_ON(!inno_vga);
	if (inno_vga->chip.irq_enable)
		return inno_vga->chip.irq_enable(&inno_vga->chip, flags);
}

void inno_vga_hw_irq_disable(struct vga_device_t *inno_vga, unsigned int flags)
{
	BUG_ON(!inno_vga);
	if (inno_vga->chip.irq_disable)
		return inno_vga->chip.irq_disable(&inno_vga->chip, flags);
}

static int inno_vga_hw_init(struct vga_device_t *inno_vga)
{
	int retcode = 0;

	if (inno_vga->chip.hw_init)
		return inno_vga->chip.hw_init(&inno_vga->chip);

	return retcode;
}

static void inno_vga_hw_fini(struct vga_device_t *inno_vga)
{
	if (inno_vga->chip.hw_fini)
		inno_vga->chip.hw_fini(&inno_vga->chip);
}

static int innodpu_vga_zoomenable_id(int vga_id)
{
	int zoom_id = -1;

	switch (vga_id) {
		case 0:
			zoom_id = VGA_ZOOM_ENABLE;
			break;
		default:
			break;
	}
	return zoom_id;
}

static int inno_vga_chip_init(struct vga_device_t *inno_vga)
{
	chip_type_e plat;
	int retcode = 0;
	int zoom_id = 0;

	zoom_id = innodpu_vga_zoomenable_id(inno_vga->vga_id);
	if ((zoom_id >= 0) && (hal_zoom_is_enable(inno_vga->dev->parent, zoom_id) == 1)) {
		inno_vga->chip.hal_zoom_enable = true;
		inno_vga->zoom_id = zoom_id;
	} else {
		inno_vga->chip.hal_zoom_enable = false;
		inno_vga->zoom_id = 0;
	}

	inno_vga->chip.hal_edid_mode = hal_vga_edid_mode(inno_vga->dev->parent, inno_vga->vga_id);
	if (inno_vga->chip.hal_edid_mode < EDID_AUTO_READ) {
		inno_vga->chip.hal_edid_mode = EDID_AUTO_READ;
	}

	innodpu_info(inno_vga->dev, DPU_UT_INIT, "hal_vga-%d edid_mode %d\n",
			inno_vga->vga_id, inno_vga->chip.hal_edid_mode);

	plat = hal_get_chiptype(inno_vga->dev->parent);
	switch(plat) {
	case CHIP_G1P_SOC:
		innodpu_info(inno_vga->dev, DPU_UT_INIT, "%s start init g1p soc.\n", inno_vga->name);
		retcode = g1p_soc_vga_chip_init(&inno_vga->chip,
				inno_vga->dev, inno_vga->vga_id, inno_vga->drm_dev);
		inno_vga->chip.hal_zoom_enable = false;
		break;
	case CHIP_G0M_SOC:
		innodpu_info(inno_vga->dev, DPU_UT_INIT, "%s start init g0m soc.\n", inno_vga->name);
		inno_vga->pmbus_adapter = hal_get_pmbus_adapter(inno_vga->dev->parent, 0);
		retcode = g0m_soc_vga_chip_init(&inno_vga->chip,
				inno_vga->dev, inno_vga->vga_id, inno_vga->drm_dev);
		break;
	default:
		innodpu_err(inno_vga->dev, "%s does not currently support %d platform.\n",
			inno_vga->name, plat);
		retcode = -EINVAL;
		break;
	}

	return retcode;
}

static void inno_vga_chip_fini(struct vga_device_t *inno_vga)
{
	chip_type_e plat;

	plat = hal_get_chiptype(inno_vga->dev->parent);
	switch(plat) {
	case CHIP_G1_SOC:
		g1p_soc_vga_chip_fini(&inno_vga->chip);
		break;
	case CHIP_G0M_SOC:
		g0m_soc_vga_chip_fini(&inno_vga->chip);
		break;
	default:
		innodpu_err(inno_vga->dev, "%s does not currently support %d platform.\n",
			inno_vga->name, plat);
		break;
	}

	return;
}


/**
 * inno_vga_bind - innosilicon vga-driver initialization function
 * @dev: vga_device_t_info allocated when inno_vga_device_t_register
 * parent(dev->parent) is &pci_dev.dev
 * @master: component master
 * @data: point of struct drm_device
 *
 * This function initializes the innosilicon vga device
 * 1. alloc and init vga handle
 * 2. vga hardware init
 * 3. vga ddcci supports
 * 4. creation and bonding of HDMI connectors and encoders
 * 5. setup vga irq
 *
 * Returns:
 * 0 if it is OK, errno otherwise.
 */
static int inno_vga_bind(struct device *dev,
					  struct device *master, void *data)
{
	int retcode = 0;
	struct drm_device * drm_dev = data;
	struct vga_device_t *inno_vga = NULL;
	plat_data_t *pdata =  dev_get_platdata(dev);

	innodpu_info(dev, DPU_UT_INIT,"start\n");
	BUG_ON(!dev);
	BUG_ON(!data);

	inno_vga = devm_kmalloc(dev, sizeof(*inno_vga), GFP_KERNEL);
	if (!inno_vga) {
		innodpu_err(dev, "Alloc vga-%d handle failed. Short of memory.\n",
			pdata->dev_idx);
		return -ENOMEM;
	}
	memset(inno_vga, 0, sizeof(*inno_vga));
	inno_vga->vga_id = pdata->dev_idx;
	atomic_set(&inno_vga->adapt_cnt, 0);
	inno_vga->dev = get_device(dev);
	inno_vga->drm_dev = drm_dev;
	inno_vga->name = kasprintf(GFP_KERNEL, "inno-vga-%d",
		inno_vga->vga_id);
	if (!inno_vga->name) {
		innodpu_err(dev, "Alloc vga-%d name failed. Short of memory.\n",
			pdata->dev_idx);
		retcode = -ENOMEM;
		goto err_out_name;
	}
	dev_set_drvdata(dev, inno_vga);
	innodpu_info(dev, DPU_UT_INIT, "%s start bind.\n", inno_vga->name);

	retcode = inno_vga_chip_init(inno_vga);
	if (retcode) {
		innodpu_err(dev, "%s chip Init failed-%d.\n", inno_vga->name, retcode);
		goto err_chip_init;
	}

	retcode = inno_vga_hw_init(inno_vga);
	if (retcode) {
		innodpu_err(dev, "%s hw Init failed-%d.\n", inno_vga->name, retcode);
		goto err_hw_init;
	}

	retcode = inno_vga_connector_attach_encoder(inno_vga);
	if (retcode) {
		innodpu_err(dev, "%s connector and encoder Init failed-%d.\n",
			inno_vga->name, retcode);
		goto err_attach_encoder;
	}

	inno_vga->auto_setup_wq = inno_create_singlethread_wq("vga auto_calib_wq");
	if (!inno_vga->auto_setup_wq) {
		retcode = -1;
		innodpu_err(dev, "%s auto_setup_sw failed-%d.\n", inno_vga->name, retcode);
		goto err_attach_encoder;
	}
	INIT_DELAYED_WORK(&inno_vga->auto_setup_work, inno_vga_auto_setup_execute);

	INIT_LIST_HEAD(&inno_vga->auto_setup_list);
	inno_vga->auto_setup_mutex = inno_mutex_alloc();
	inno_vga->adapt_lock = inno_mutex_alloc();
	if (!inno_vga->auto_setup_mutex || !inno_vga->adapt_lock) {
		retcode = -1;
		innodpu_err(dev, "mutex alloc failed.\n");
		goto err_attach_encoder;
	}

	inno_vga->adapt_wait = inno_waitqueue_head_alloc();
	if (!inno_vga->adapt_wait) {
		retcode = -1;
		innodpu_err(dev, "vga adapt wait alloc failed\n");
		goto err_attach_encoder;
	}
	inno_init_waitqueue_head(inno_vga->adapt_wait);

	innodpu_info(dev, DPU_UT_INIT,"end\n");
	return retcode;

err_attach_encoder:
	if (inno_vga->adapt_wait) {
		inno_waitqueue_head_free(inno_vga->adapt_wait);
		inno_vga->adapt_wait = NULL;
	}

	if (inno_vga->auto_setup_mutex)
		inno_mutex_free(inno_vga->auto_setup_mutex);
	if (inno_vga->adapt_lock)
		inno_mutex_free(inno_vga->adapt_lock);
	inno_vga_hw_fini(inno_vga);
err_hw_init:
	inno_vga_chip_fini(inno_vga);
err_chip_init:
	if (inno_vga->name)
		kfree(inno_vga->name);
err_out_name:
	put_device(dev);
	devm_kfree(dev, inno_vga);

	return retcode;
}

/**
* inno_vga_unbind - innosilicon vga-driver initialization function
* @dev: vga_device_t_info allocated when inno_vga_device_t_register
* parent(dev->parent) is &pci_dev.dev
* @master: component master
* @data: point of struct drm_device
*
* This function deinitializes the innosilicon vga device
*
* Once drm_dev_register is called, the connector and encoder
* resources are released by the drm_mode_config_cleanup function!!!
* So we don't need to worry about the release of the connector and encoder here.
*/
static void inno_vga_unbind(struct device *dev,
						  struct device *master, void *data)

{
	struct vga_device_t *inno_vga = NULL;

	innodpu_info(dev, DPU_UT_INIT,"start\n");
	BUG_ON(!dev);
	BUG_ON(!data);

	inno_vga = dev_get_drvdata(dev);
	if (!inno_vga) {
		innodpu_err(dev, "vga handle is NULL\n");
		return;
	}

	atomic_set(&inno_vga->adapt_cnt, 0);
	inno_wake_up_interruptible(inno_vga->adapt_wait);
	cancel_delayed_work_sync(&inno_vga->auto_setup_work);
	inno_destroy_workqueue(inno_vga->auto_setup_wq);

	inno_vga_hw_fini(inno_vga);
	inno_vga_chip_fini(inno_vga);
	kfree(inno_vga->name);

	if (!list_empty(&inno_vga->auto_setup_list))
		inno_vga_auto_setup_mode_clear(inno_vga);
	if (inno_vga->auto_setup_mutex)
		inno_mutex_free(inno_vga->auto_setup_mutex);
	if (inno_vga->adapt_lock)
		inno_mutex_free(inno_vga->adapt_lock);

	if (inno_vga->adapt_wait) {
		inno_waitqueue_head_free(inno_vga->adapt_wait);
		inno_vga->adapt_wait = NULL;
	}

	put_device(dev);
	devm_kfree(dev, inno_vga);
	innodpu_info(dev, DPU_UT_INIT,"end\n");
}


static const struct component_ops s_inno_vga_ops = {
	.bind = inno_vga_bind,
	.unbind = inno_vga_unbind,
};

static int inno_vga_probe(struct platform_device *pdev)
{
	BUG_ON(!pdev);
	return component_add(&pdev->dev, &s_inno_vga_ops);
}

static int inno_vga_remove(struct platform_device *pdev)
{
	BUG_ON(!pdev);
	component_del(&pdev->dev, &s_inno_vga_ops);
	return 0;
}

static struct platform_device_id s_inno_vga_device_id_table[] = {
	{.name = INNO_VGA_DEVICE_NAME, .driver_data = 0},
	{},
};
MODULE_DEVICE_TABLE(platform, s_inno_vga_device_id_table);

static int inno_vga_suspend(struct device *dev)
{
	struct vga_device_t *inno_vga = dev_get_drvdata(dev);

	innodpu_info(dev, DPU_UT_INIT,"suspend\n");

	if (inno_vga) {
		atomic_set(&inno_vga->adapt_cnt, 0);
		inno_wake_up_interruptible(inno_vga->adapt_wait);
		cancel_delayed_work_sync(&inno_vga->auto_setup_work);
		inno_vga_hw_fini(inno_vga);
		if (!list_empty(&inno_vga->auto_setup_list))
			inno_vga_auto_setup_mode_clear(inno_vga);
	}

	return 0;
}

static int inno_vga_resume(struct device *dev)
{
	struct vga_device_t *inno_vga = dev_get_drvdata(dev);

	innodpu_info(dev, DPU_UT_INIT,"resume\n");
	if (inno_vga) {
		inno_vga_hw_init(inno_vga);
	}

	return 0;
}

static void inno_vga_shutdown(struct platform_device *pdev)
{
	struct vga_device_t *inno_vga = inno_platform_get_drvdata(pdev);

	if (inno_vga) {
		inno_vga_hw_fini(inno_vga);
	}
}

static const struct dev_pm_ops inno_vga_pm_ops = {
	SET_SYSTEM_SLEEP_PM_OPS(inno_vga_suspend, inno_vga_resume)
};

struct platform_driver g_innogpu_vga_driver = {
	.probe = inno_vga_probe,
	.remove = inno_vga_remove,
	.shutdown = inno_vga_shutdown,
	.driver = {
		.name = INNO_VGA_DEVICE_NAME,
		.pm = &inno_vga_pm_ops,
	},
	.id_table = s_inno_vga_device_id_table,
};
